package quimufu.colourful_portals.portal;

import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.entity.Entity;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.DyeColor;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import quimufu.colourful_portals.config.ColourfulPortalConfig;
import quimufu.colourful_portals.general_util.LinkedList;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static quimufu.colourful_portals.ColourfulPortalsMod.*;

public class PortalManager {

    private final PortalListComponent portalCandidateList;
    private final PortalListComponent portalList;

    public PortalLinkingSystem getLinkingSystem() {
        return linkingSystem;
    }

    private final PortalLinkingSystem linkingSystem;

    public PortalManager(PortalLinkingSystem portalLinkingSystem, PortalListComponent portalCandidateList, PortalListComponent portalList) {
        this.portalCandidateList = portalCandidateList;
        this.portalList = portalList;
        this.linkingSystem = portalLinkingSystem;
    }

    public void onPortalBlockPlaced(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockPlaced");
        LOGGER.debug("onPortalBlockPlaced, {}", blockId);

        Identifier dimIdentifier = PortalHelper.getDimId(world);

        //delete portalCandidates obstructed by PortalBlocks for consistency
        List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dimIdentifier);
        LOGGER.debug("got containing PortalCandidates , {}", portalCandidates);
        for (PortalRepresentation portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate, blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);

                portalCandidateList.removePortal(blockId, portalCandidate);
            }
        }
        //find new portalCandidates created by PortalBlock placement
        List<PortalRepresentation> portals = PortalHelper.findPortalCandidates(world, pos, blockId);
        LOGGER.debug("new portalCandidates found, {}", portals);

        for (PortalRepresentation portal : portals) {
            portalCandidateList.createPortal(blockId, portal);
        }
        world.getProfiler().pop();
    }

    public void onPortalBlockBroken(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockBroken");
        LOGGER.debug("onPortalBlockBroken, {}", blockId);

        //check portalCandidate validity
        Identifier dimensionIdentifier = PortalHelper.getDimId(world);
        List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dimensionIdentifier);
        LOGGER.debug("getContainingPortals, {}", portalCandidates);
        for (PortalRepresentation portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate, blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);
                portalCandidateList.removePortal(blockId, portalCandidate);
            }
        }

        //add portalCandidates deobstructed by PortalBlock removal
        for (Direction direction : Direction.values()) {
            List<PortalRepresentation> newPortalCandidates = PortalHelper.findPortalCandidates(world, pos.add(direction.getVector()), blockId);

            for (PortalRepresentation newPortal : newPortalCandidates) {
                LOGGER.debug("potentially new candidate after deobstruction , {}", newPortal);
                portalCandidateList.createPortal(blockId, newPortal);
            }
        }


        //check portal validity
        List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dimensionIdentifier);
        boolean changed = false;
        for (PortalRepresentation portal : portals) {
            if (!PortalHelper.isValidPortal(world, portal, blockId)) {
                LOGGER.debug("portal became invalid ,{} {}", dimensionIdentifier, portal);

                portalList.removePortal(blockId, portal);
                linkingSystem.unLinkPortal(portal);
                removePortalBlocks(world, portal.location());
                changed = true;
            }

        }
        if (changed) {
            fixPortalGroup(blockId, portalList, world.getServer());
        }

        world.getProfiler().pop();
    }

    private void removePortalBlocks(ServerWorld world, BlockBox portal) {
        BlockPos.stream(portal)
                .filter(blockPos -> world.getBlockState(blockPos).isOf(PORTAL_BLOCK))
                .forEach(blockPos -> world.setBlockState(blockPos, Blocks.AIR.getDefaultState()));
    }

    private void fixPortalGroup(Identifier blockId, PortalListComponent portalList, MinecraftServer server) {
        LinkedList<PortalRepresentation> portalsWithDim = portalList.getPortals(blockId);
        if (portalsWithDim.size() == 1) {
            PortalRepresentation portalRepresentation = portalsWithDim.getFirst();
            ServerWorld world = PortalHelper.getPortalWorld(server, portalRepresentation);
            if (world == null) {
                LOGGER.error("error fixing portalGroup, dimensionId {} was null", portalRepresentation.dimensionId());
                return;
            }
            portalList.removePortal(blockId, portalRepresentation);
            linkingSystem.unLinkPortal(portalRepresentation);
            removePortalBlocks(world, portalRepresentation.location());
            return;
        }
        for (PortalRepresentation portalRepresentation : portalsWithDim) {
            assurePortalBlocksPlaced(blockId, portalRepresentation, server);
        }
        linkingSystem.linkPortals(portalsWithDim);

    }

    private void assurePortalBlocksPlaced(Identifier blockId, PortalRepresentation portalRepresentation, MinecraftServer server) {
        DyeColor color = ColourfulPortalConfig.colorOf(blockId.toString());
        if (color != null) {
            Direction.Axis portalPlaneAxis = PortalHelper.getAxisW(portalRepresentation)
                    .rotateClockwise(Direction.Axis.Y)
                    .getAxis();
            ServerWorld portalWorld = PortalHelper.getPortalWorld(server, portalRepresentation);
            if (portalWorld == null) {
                LOGGER.error("error placing portalRepresentation planes, dimensionId {} was null", portalRepresentation.dimensionId());
                return;
            }
            BlockState portalBlockState = PORTAL_BLOCK.getStateWith(color, portalPlaneAxis);

            PortalHelper.insideOf(portalRepresentation)
                    .forEachRemaining(blockPos -> {
                        BlockState blockState = portalWorld.getBlockState(blockPos);
                        if (blockState.isAir() || blockState.getFluidState().isOf(PORTAL_FLUID)) {
                            portalWorld.setBlockState(blockPos, portalBlockState);
                        } else if (blockState.isReplaceable()) {
                            portalWorld.breakBlock(blockPos, true);
                            portalWorld.setBlockState(blockPos, portalBlockState);
                        }
                    });
        }

    }


    public boolean tryIgnite(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal ignition");
        LOGGER.info("portal ignition");
        Set<Identifier> blockIds = portalCandidateList.getBlockIds();
        boolean ret = false;
        for (Identifier blockId : blockIds) {
            Identifier dim = PortalHelper.getDimId(world);
            List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation current : portalCandidates) {
                if (PortalHelper.isValidPortal(world, current, blockId)) {
                    PortalRepresentation next = portalCandidateList.getNext(blockId, current);
                    ServerWorld nextWorld = PortalHelper.getPortalWorld(world.getServer(), next);
                    if (!next.equals(current)
                        && (!portalList.containsPortal(blockId, current) || !portalList.containsPortal(blockId, next))
                        && PortalHelper.isPortalPlaceable(nextWorld, next, blockId)) {
                        portalList.createPortal(blockId, current);
                        portalList.createPortal(blockId, next);
                        ret = true;
                    }
                    fixPortalGroup(blockId, portalList, world.getServer());
                }
            }
        }
        world.getProfiler().pop();
        return ret;
    }

    public boolean canExtend(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal extension check");
        Set<Identifier> blockIds = portalList.getBlockIds();
        Identifier dim = PortalHelper.getDimId(world);

        for (Identifier blockId : blockIds) {
            List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation portal : portals) {
                if (PortalHelper.isValidPortal(world, portal, blockId)) {
                    PortalRepresentation last = portalList.getLast(blockId);
                    PortalRepresentation next = portalCandidateList.getNext(blockId, last);
                    ServerWorld nextWorld = PortalHelper.getPortalWorld(world.getServer(), next);

                    if (!portalList.containsPortal(blockId, next)
                        && PortalHelper.isPortalPlaceable(nextWorld, next, blockId)) {
                        world.getProfiler().pop();
                        return true;
                    }
                }
            }

        }

        world.getProfiler().pop();
        return false;
    }

    public boolean extend(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal extension");
        Set<Identifier> blockIds = portalList.getBlockIds();
        Identifier dim = PortalHelper.getDimId(world);

        boolean ret = false;
        for (Identifier blockId : blockIds) {
            List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation portal : portals) {
                if (PortalHelper.isValidPortal(world, portal, blockId)) {
                    PortalRepresentation last = portalList.getLast(blockId);
                    PortalRepresentation next = portalCandidateList.getNext(blockId, last);
                    ServerWorld nextWorld = PortalHelper.getPortalWorld(world.getServer(), next);

                    if (!portalList.containsPortal(blockId, next)
                        && PortalHelper.isPortalPlaceable(nextWorld, next, blockId)) {
                        portalList.createPortal(blockId, next);
                        ret = true;
                    }
                    fixPortalGroup(blockId, portalList, world.getServer());
                }
            }

        }

        world.getProfiler().pop();
        return ret;
    }

    public void onPortalPassed(Entity entity, BlockPos pos, ServerWorld world, Direction.Axis a) {
        linkingSystem.onPortalPassed(entity, pos, world, a);
    }

    public boolean onMovementThroughBlock(Entity entity, BlockPos pos, ServerWorld world) {
        return linkingSystem.movementCallback(entity, pos, world);
    }

    public void onLoad(MinecraftServer minecraftServer) {
        HashSet<Identifier> blockIdsToDelete = new HashSet<>(portalCandidateList.getBlockIds());
        blockIdsToDelete.addAll(portalList.getBlockIds());
        blockIdsToDelete.removeAll(PORTAL_BLOCKS);
        for (Identifier blockId : blockIdsToDelete) {
            for (PortalRepresentation portal : portalCandidateList.getPortals(blockId)) {
                portalCandidateList.removePortal(blockId, portal);
            }
            for (PortalRepresentation portal : portalList.getPortals(blockId)) {
                portalList.removePortal(blockId, portal);
                linkingSystem.unLinkPortal(portal);
                ServerWorld portalWorld = PortalHelper.getPortalWorld(minecraftServer, portal);
                if (portalWorld != null) {
                    removePortalBlocks(portalWorld, portal.location());
                }
            }

        }

    }
}
